package me.topchetoeu.revengmc.core;

import java.io.IOException;
import java.net.Proxy;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;

import javax.annotation.Nullable;

import com.mojang.authlib.GameProfile;
import com.mojang.serialization.Lifecycle;

import net.minecraft.CrashReport;
import net.minecraft.SystemReport;
import net.minecraft.core.MappedRegistry;
import net.minecraft.core.Registry;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.server.level.progress.LoggerChunkProgressListener;
import net.minecraft.server.players.PlayerList;
import net.minecraft.util.datafix.DataFixers;
import net.minecraft.world.Difficulty;
import net.minecraft.world.level.DataPackConfig;
import net.minecraft.world.level.GameRules;
import net.minecraft.world.level.GameType;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.LevelSettings;
import net.minecraft.world.level.biome.Biome;
import net.minecraft.world.level.dimension.DimensionType;
import net.minecraft.world.level.levelgen.NoiseGeneratorSettings;
import net.minecraft.world.level.levelgen.WorldGenSettings;
import net.minecraft.world.level.storage.LevelStorageSource;
import net.minecraft.world.level.storage.PrimaryLevelData;
import net.minecraft.world.level.storage.LevelStorageSource.LevelStorageAccess;

// God i love this word. Perfectly sums up Notch's coding skills
public class BogusServer extends MinecraftServer {
	private static final LevelSettings BOGUS_SETTINGS = new LevelSettings("Bogus Level", GameType.CREATIVE, false, Difficulty.NORMAL, true, new GameRules(), DataPackConfig.DEFAULT);

	public final ServerLevel level;

	public BogusServer(MinecraftContext ctx, long seed) throws IOException {
		this(
			ctx, seed, LevelStorageSource.createDefault(Path.of("levels")),
			ctx.registry().registryOrThrow(Registry.NOISE_GENERATOR_SETTINGS_REGISTRY),
			ctx.registry().registryOrThrow(Registry.BIOME_REGISTRY),
			ctx.registry().registryOrThrow(Registry.DIMENSION_TYPE_REGISTRY)
		);
	}

	@SuppressWarnings("null")
	private BogusServer(
		MinecraftContext ctx,
		long seed, LevelStorageSource levelStorageSrc,
		Registry<NoiseGeneratorSettings> noiseGenerators,
		Registry<Biome> biomes,
		Registry<DimensionType> dimensionTypes
	) throws IOException {
		super(
			null, ctx.registry(), levelStorageSrc.createAccess("world-" + seed),
			new PrimaryLevelData(
				BOGUS_SETTINGS,
				new WorldGenSettings(
					seed,
					true, false,
					WorldGenSettings.withOverworld(
						dimensionTypes,
						new MappedRegistry<>(Registry.LEVEL_STEM_REGISTRY, Lifecycle.experimental()),
						WorldGenSettings.makeDefaultOverworld(biomes, noiseGenerators, seed)
					)
				),
				Lifecycle.stable()
			),
			ctx.packRepo(),
			Proxy.NO_PROXY, DataFixers.getDataFixer(),
			ctx.resources(), null, null, null,
			LoggerChunkProgressListener::new
		);


		this.setPlayerList(new PlayerList(this, this.registryHolder, this.playerDataStorage, 1) {});
		this.level = new ServerLevel(
			this, r -> r.run(), this.levelStorage(),
			this.getWorldData().overworldData(),
			Level.OVERWORLD, ctx.overworldLocation(),
			new LoggerChunkProgressListener(0),
			this.getWorldData().worldGenSettings().overworld(),
			false, seed, List.of(), false
		);
	}

	public LevelStorageAccess levelStorage() { return this.storageSource; }

	@Override public boolean initServer() throws IOException {
		// this.createLevels(new LoggerChunkProgressListener(0));
		return true;
	}
	@Override public SystemReport fillServerSystemReport(@Nullable SystemReport report) { return report; }
	@Override public void onServerCrash(@Nullable CrashReport report) { System.exit(1); }
	@Override public boolean isHardcore() { return false; }
	@Override public int getOperatorUserPermissionLevel() { return 0; }
	@Override public int getFunctionCompilationLevel() { return 4; }
	@Override public boolean shouldRconBroadcast() { return false; }
	@Override public boolean isDedicatedServer() { return false; }
	@Override public int getRateLimitPacketsPerSecond() { return 0; }
	@Override public boolean isEpollEnabled() { return false; }
	@Override public boolean isCommandBlockEnabled() { return true; }
	@Override public boolean isPublished() { return false; }
	@Override public boolean shouldInformAdmins() { return false; }
	@Override public boolean isSingleplayerOwner(@Nullable GameProfile profile) { return false; }
	@Override public Optional<String> getModdedStatus() { return Optional.empty(); }
	@SuppressWarnings("null")
	@Override public void close() {
		try {
			if (getConnection() != null) getConnection().stop();
			getPlayerList().removeAll();
	
			for (var level : this.getAllLevels()) {
				if (level != null) level.close();
			}
	
			if (this.getSnooper().isStarted()) {
				this.getSnooper().interrupt();
			}
	
			this.storageSource.close();
		}
		catch (RuntimeException e) { throw e; }
		catch (Throwable e) { throw new RuntimeException(e); }
	}
	public void close(boolean full) {
		if (full) super.close();
		else this.close();
	}
}